/*
 * This file is part of SOCKMAND.
 * 
 * SOCKMAND is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * SOCKMAND is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with SOCKMAND.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */


#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <arpa/inet.h>
#include <netinet/in.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>

struct ipheader {
    unsigned char ip_hl:4, ip_v:4; 
    unsigned char ip_tos;
    unsigned short int ip_len;
    unsigned short int ip_id;
    unsigned short int ip_off;
    unsigned char ip_ttl;
    unsigned char ip_p;
    unsigned short int ip_sum;
    unsigned int ip_src;
    unsigned int ip_dst;
}; 


struct udpheader {
    unsigned short int uh_sport;
    unsigned short int uh_dport;
    unsigned short int uh_len;
    unsigned short int uh_check;
}; 


void build_ip_packet(struct ipheader *ip, int protocol, unsigned int srcaddr, unsigned int destaddr, int buflen){
    ip->ip_hl = 5; 
    ip->ip_v = 4; 
    ip->ip_tos = 16; 
    ip->ip_len = buflen; 
    ip->ip_id = htons(54321); 
    ip->ip_ttl = 64; 
    ip->ip_p = protocol;
    ip->ip_src = srcaddr;
    ip->ip_dst = destaddr;
}

void build_udp_header(struct udpheader *udp, unsigned short srcprt, unsigned short destport, int buflen){    
    udp->uh_sport = srcprt;
    udp->uh_dport = destport;    
    udp->uh_len = htons(buflen-(sizeof(struct ipheader)));
}

int main(int argc, char **argv){
    struct timeval start, prev, now, end;
    long elapsed_utime;    
    long elapsed_seconds;  
    long elapsed_useconds; 
    
    if(argc != 7){
        fprintf(stderr, "Usage: %s <MCAP IP address> <destination IP address> <dest port number> <number of seconds> <packets per seconds> <packetsize>\n", argv[0]);
        return 0;
    }
    
    unsigned int  mcapaddr = inet_addr(argv[1]);
    if(mcapaddr==-1){
        printf("Could not parse <MCAP IP address>\n");
        return 0;
    }
    unsigned int toaddr = inet_addr(argv[2]);
    if(toaddr==-1){
        printf("Could not parse <destination IP address>\n");
        return 0;
    }
    
    int destport=atoi(argv[3]);
    
    if(destport==-1){
        printf("Could not parse <dest port number>\n");
        return 0;
    }
    
    int srcport=9999;
    
    int seconds = atoi(argv[4]);
    if(seconds==-1){
        printf("Could not parse <number of seconds>\n");
        return 0;
    }
    int packetrate =  atoi(argv[5]);
    
    if(packetrate==-1){
        printf("Could not parse <packets per seconds>\n");
        return 0;
    }
    
    int buflen = atoi(argv[6]);
    
     if(buflen==-1){
        printf("Could not parse <packetsize>\n");
        return 0;
    }
    
    int totalpackets = packetrate*seconds;
    
    char buf[buflen];
    
    
    long utimebetweensend = 1000000/packetrate;
    
    
    printf("Sending %d packets\n", totalpackets);
    printf("%d packets / second\n", packetrate);
    printf("Sending for a total of %d seconds\n", seconds);
    printf("Microseconds between each packet %ld\n", utimebetweensend);
    
    
    struct ipheader *ip = (struct ipheader *) buf; 
    struct udpheader *udp = (struct udpheader *) (ip + 1);
    
    
    build_ip_packet(ip, 
                    IPPROTO_UDP, 
                    mcapaddr, 
                    toaddr,
                    buflen);
    
    build_udp_header(udp, htons(srcport), htons(destport), buflen);
    
    
    int tunnel_sockfd = socket(AF_INET,SOCK_DGRAM,0);
    
    struct sockaddr_in servaddr;    
    bzero(&servaddr,sizeof(servaddr));
    servaddr.sin_family = AF_INET;
    servaddr.sin_addr.s_addr = htonl(INADDR_ANY);
    servaddr.sin_port = htons(5000);
    
    bind(tunnel_sockfd,(struct sockaddr *)&servaddr,sizeof(servaddr));
    
    bzero(&servaddr,sizeof(servaddr));
    servaddr.sin_family = AF_INET;
    
    servaddr.sin_addr.s_addr = mcapaddr;
    servaddr.sin_port = htons(5000);
    
    long left = 0;
    int sendtosize = sizeof(servaddr);
    
    int i;
    gettimeofday(&start, NULL);
    sendto(tunnel_sockfd, buf, buflen, 0, (struct sockaddr *)&servaddr,sendtosize);            
    
    gettimeofday(&prev, NULL);
    
    
    for(i=0;i<totalpackets-1;i++)
    {
        while(1){
            gettimeofday(&now, NULL);
            elapsed_seconds  = now.tv_sec  - prev.tv_sec;
            elapsed_useconds = now.tv_usec - prev.tv_usec;
            elapsed_utime = (elapsed_seconds) * 1000000 + elapsed_useconds;
            left = elapsed_utime - utimebetweensend;
            if(left > 0)
                break;
        }
        sendto(tunnel_sockfd, buf, buflen, 0, (struct sockaddr *)&servaddr,sendtosize);
        memcpy(&prev,&now,sizeof(struct timeval)); 
        prev.tv_usec -= left;
    }
    
    gettimeofday(&end, NULL);
    
    elapsed_seconds  = end.tv_sec  - start.tv_sec;
    elapsed_useconds = end.tv_usec - start.tv_usec;
    elapsed_utime = (elapsed_seconds) * 1000000 + elapsed_useconds;
    
    double secs = elapsed_utime/1000000.0;
    
    printf("Sent %d packets\n", totalpackets);
    double actualrate = (double)totalpackets/secs; 
    printf("Actual packets / second %f\n",actualrate);
    printf("Elapsed time = %ld microseconds\n", elapsed_utime);
    printf("Elapsed time = %f seconds\n", secs);
    close(tunnel_sockfd);
    return 0;
}