/*
 * This file is part of SOCKMAND.
 * 
 * SOCKMAND is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * SOCKMAND is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with SOCKMAND.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */


#include <netinet/ip.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <arpa/inet.h>
#include <signal.h>
#include <sys/select.h>
#include <unistd.h>

#include "core.h"
#include "sockmandapi.h"
#include "libpcaphandler.h"
#include "rawsockhandler.h"
#include "tunnel.h"
#include "ipcont.h"
#include "headers.h"
#include "udpcont.h"
#include "tcpcont.h"
#include "utils.h"
#include "signalhandler.h"
#include "migapi.h"
#include "ui.h"
#include "eventhandler.h"


int daemonized = 0;

int core_exit(){
    printf("Exiting SOCKMAND\n");
    tunnel_exit();
    libpcaphandler_exit();
    rawsock_exit();
    signal_exit();
    appapi_exit();
    return 0;
}

int core_init(int isdaemonized, unsigned int default_mcap){
    daemonized = isdaemonized;
    printf("SOCKMAND init\n");
    get_my_addr();
    signal(SIGINT, &signal_handler);
    migapi_init(default_mcap);
    tunnel_init();
    libpcaphandler_init();
    rawsock_init();
    signalinit();
    appapi_init();
    handle_events();
    return 0;
}

//Receieves a ip packet from ip controller
int handle_ip_packet(struct ipheader *ippacket){
    if(is_my_address(ippacket->ip_src)){
        return write_ip_packet(ippacket);
    }
    else{
        return write_tunneled_packet(ippacket, ippacket->ip_src);
    }
}

//Receieves a ip packet from the tunnel
void handle_tunneled_packet(struct ipheader *ippacket, unsigned int src_address){
    if(is_app_present(ippacket)){
        read_ip_packet(ippacket);  
    }
    else{
        if(is_my_address(ippacket->ip_src)){
            write_ip_packet(ippacket);
        }
        else{
            //Major error
            printf("Unexpected packet in handle tunneled packet\n");
        }
    }
}

//Receieves a captured ip packet from the libpcaphandler
void handle_captured_packet(struct ipheader *ippacket){
    if(is_app_present(ippacket)){
        read_ip_packet(ippacket);
    }
    else{ //App somewhere else
        write_tunneled_packet(ippacket, get_app_location(ippacket));
    }
}

struct connectioninfo *get_connection_info_by_uuid(int uuid){
    struct connectioninfo * ret = udp_get_connection_info_by_uuid(uuid);
    if(ret == 0)
        ret = tcp_get_connection_info_by_uuid(uuid);
    return ret;
}

void send_socket_state(struct connectioninfo *c, unsigned int address){
  
    int cblen = 0;
    if(c->protocol == IPPROTO_TCP)
        cblen = sizeof(struct tcp_cb);
    else if(c->protocol == IPPROTO_UDP)
        cblen = sizeof(struct udp_cb);
    
    int unackedpacketslength = 0;
    if(c->protocol == IPPROTO_TCP && address == c->current_address){
        unackedpacketslength = get_serialized_timeout_length((struct tcp_cb*)c);
    }
    
    
    int length = sizeof(struct socketstate) + cblen + unackedpacketslength;
    
    struct socketstate* ss = malloc(length);        
    ss->signaltype = SOCKETSTATE;
    memcpy(ss+1, c, cblen);
    ss->protocol = c->protocol; 
    ss->length = length;


    
    if(unackedpacketslength != 0){
        char* timeouts = ((char*)ss) + cblen;
        serialize_timeouts(timeouts, get_cb_by_pointer(c));

    }
    
    sendsignal((char*)ss, length, address);
    free(ss);
}

int migrate_socket(int uuid, unsigned int newaddr){
    if(newaddr == get_my_addr()){
        printf("Migrating to here, not possible\n");
        return -1;
    }
    
    struct connectioninfo *c = get_connection_info_by_uuid(uuid);
    if(c==0){
        printf("Unknown uuid\n");
        return -1;
    }
    
    c->current_address = newaddr;
    
    if(c->mcap_address != get_my_addr() && c->mcap_address != newaddr){
        //Send signal to the MCAP
        struct signalheader s;
        s.signaltype = MIGRATING_TO;
        s.newaddress = newaddr;
        s.uuid = uuid;
        sendsignal((char*)&s, sizeof(struct signalheader), c->mcap_address);
    }
    
    send_socket_state(c, newaddr);
    c->status = MIGRATEDAWAY;
    c->sockmandapi_id = -1;
    return 0;
}


void import_socket_state(struct socketstate* ss){
    struct connectioninfo *c = (struct connectioninfo *) (ss+1);
    
    c->sockmandapi_id = -1; //No app has connected here yet
    
    if(c->current_address != get_my_addr()){ 
        c->status = MIGRATEDAWAY; //This node will be the anchornode
        fprintf(stderr,"Acting as MCAP for socket %d\n", c->uuid);
    }
    else{
        fprintf(stderr,"Socket %d migrated here\n", c->uuid);
    }
    
    
    int cblen = 0;
    if(c->protocol == IPPROTO_TCP)
        cblen = sizeof(struct tcp_cb);
    else if(c->protocol == IPPROTO_UDP)
        cblen = sizeof(struct udp_cb);
    
    
    struct connectioninfo *oldinfo = get_connection_info_by_uuid(c->uuid);
    if(oldinfo!=0){
        //The socket has either use this node as an MCAP or has been here before
        memcpy(oldinfo, c, cblen);
        if(c->protocol == IPPROTO_TCP){
            int timeoutslen = ss->length - sizeof(struct tcp_cb) - sizeof(struct socketstate);
            if(timeoutslen>0){

                char* timeouts = (char* )ss + sizeof(struct tcp_cb);
                deserialize_timeouts(timeouts, get_cb_by_pointer(oldinfo));
            }
        }
    }
    else{
        if(c->protocol == IPPROTO_TCP){
            int cb = tcp_importstate((struct tcp_cb*)c);
            int timeoutslen = ss->length - sizeof(struct tcp_cb) - sizeof(struct socketstate);
            if(timeoutslen>0){
                char* timeouts = (char* )ss + sizeof(struct tcp_cb);
                deserialize_timeouts(timeouts, cb);
            }
            
        }
        else if(c->protocol == IPPROTO_UDP)
            udp_importstate((struct udp_cb*)c);
    }
}

//Handles a signal receieved over a tunnel
void handle_signal(char* signal){
    struct signalheader *s = (struct signalheader*) signal;
    struct connectioninfo *c = get_connection_info_by_uuid(s->uuid);
    
    if(s->signaltype == MIGRATING_TO){      
         c->current_address = s->newaddress;
         fprintf(stderr,"Socket with uuid %d migrated to %s\n", s->uuid, ip_to_char(s->newaddress));
    }
    else if(s->signaltype == SOCKETSTATE){
        import_socket_state((struct socketstate*) signal);
    }
    else{
        printf("Major error, unknown signal received\n");
    }
}

