/*
 * This file is part of SOCKMAND.
 * 
 * SOCKMAND is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * SOCKMAND is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with SOCKMAND.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

#include "ipcont.h"
#include "udpcont.h"
#include "headers.h"
#include "core.h"
#include "utils.h"
#include "libpcaphandler.h"
#include "sockmandapi.h"


#include <stdlib.h>
#include <string.h>
#include <arpa/inet.h>
#include <stdio.h>

int next_udpcb = 0;
struct udp_cb udp_cbs[1024];

void udp_importstate(struct udp_cb *c){
    memcpy(&udp_cbs[next_udpcb], c, sizeof(struct udp_cb));
    udp_cbs[next_udpcb].bufferedpacketsfirst = 0;
    udp_cbs[next_udpcb].bufferedpacketslast = 0;
    if(udp_cbs[next_udpcb].mcap_address == get_my_addr()){
        register_udp_capture(udp_cbs[next_udpcb].sourceport);
    }
    next_udpcb++;
}


int get_udp_cb(struct ipheader* ip){
    //takes an incoming packet as arg and returns the corresponding cb
    struct udpheader *udp = (struct udpheader *) (ip + 1);
    
    int i;
    for(i = 0; i < next_udpcb; i++){
        if(udp_cbs[i].status != UDPCLOSED){
            if(ip->ip_dst == udp_cbs[i].mcap_address &&
                udp->uh_dport == udp_cbs[i].sourceport)
                return i;            
        }
    }
    return -1;
}

int udp_is_app_present(struct ipheader *ip){
    int cb = get_udp_cb(ip);
    if(cb == -1)
        return 0;
    return is_my_address(udp_cbs[cb].current_address);
}

unsigned int udp_get_app_location(struct ipheader *ip){
    return udp_cbs[get_udp_cb(ip)].current_address;
}

void* get_udp_cb_pointer(int id){
    return &udp_cbs[id];
}

int udp_get_cb_by_uuid(int uuid){
    int i;
    for(i = 0;i < next_udpcb; i++){
        if(udp_cbs[i].uuid == uuid)
            return i;
    }
    return -1;
}


void* udp_get_connection_info_by_uuid(int uuid){
    int i;
    for(i = 0;i < next_udpcb; i++){
        if(udp_cbs[i].uuid == uuid)
            return get_udp_cb_pointer(i);
    }
    return 0;
}


void set_udp_unbound(int cb){
    if(udp_cbs[cb].status == UDPACTIVE && udp_cbs[cb].sockmandapi_id != -1)
        udp_cbs[cb].status = UDPCLOSED;
}

void build_udp_header(struct udpheader *udp, unsigned short srcprt, unsigned short destport, int len){    
    udp->uh_sport = srcprt;
    udp->uh_dport = destport;    
    udp->uh_len = htons(sizeof(struct udpheader) + len);
}

int udpwrite(int cb, const void *buf, int len, unsigned int destaddr, short destport){
    if(len> MAXUDPPAYLOAD){
        printf("Major error, to large udp packet\n");
        return -1;
    }
    
    char buffer[ETHERNETMTU]; 
    bzero(&buffer, ETHERNETMTU);
    struct ipheader *ip = (struct ipheader *) buffer; 
    struct udpheader *udp = (struct udpheader *) (ip + 1);
    char *data = (char*)(udp + 1);
    
    build_ip_packet(ip, 
                    IPPROTO_UDP, 
                    udp_cbs[cb].mcap_address, 
                    destaddr, 
                    sizeof(struct udpheader) + len);
    
    build_udp_header(udp, udp_cbs[cb].sourceport, destport, len);
    
    memcpy(data, buf, len);
    send_ip_packet(ip);
    return len;
}


int udpbind(int sockmandapi_id, int uuid, short sourceport, unsigned int mcap_address){
    int i;
    for(i = 0; i < next_udpcb; i++){
        if(udp_cbs[i].status == UDPACTIVE && udp_cbs[i].sourceport == sourceport && udp_cbs[i].mcap_address == mcap_address){
            return -1;
        }
    }
    
    udp_cbs[next_udpcb].sockmandapi_id = sockmandapi_id;
    udp_cbs[next_udpcb].uuid = uuid;
    udp_cbs[next_udpcb].mcap_address = mcap_address;
    udp_cbs[next_udpcb].current_address = get_my_addr();
    udp_cbs[next_udpcb].sourceport = sourceport;
    udp_cbs[next_udpcb].status = UDPACTIVE;
    udp_cbs[next_udpcb].protocol = IPPROTO_UDP;
    udp_cbs[next_udpcb].bufferedpacketsfirst = 0;
    udp_cbs[next_udpcb].bufferedpacketslast = 0;
    
    if(is_my_address(mcap_address))
        register_udp_capture(sourceport);
    else
        send_socket_state((struct connectioninfo *)&udp_cbs[next_udpcb], mcap_address);
    
    return next_udpcb++;
}

void udp_deliverbufferdpackets(int cb){
    struct bufferedpacket* t = udp_cbs[cb].bufferedpacketsfirst;
    if(t==0)
        return;
    
    while(t!=0){
        struct ipheader* ip = t->ip;
        struct udpheader *udp = (struct udpheader *) (ip + 1);
        int payloadlen = ntohs(udp->uh_len) - sizeof(struct udpheader);
        
        deliver_udp_data( udp_cbs[cb].sockmandapi_id, 
                          (char*)(udp + 1), 
                          payloadlen, 
                          udp->uh_sport, 
                          ip->ip_src);
        
        struct bufferedpacket* tmp = t->next;
        free(ip);
        free(t);
        t=tmp;
    } 
}


void udpread(struct ipheader *ip){
    struct udpheader *udp = (struct udpheader *) (ip + 1);
    int cb = get_udp_cb(ip);
    int payloadlen = ntohs(udp->uh_len) - sizeof(struct udpheader);
    if(udp_cbs[cb].sockmandapi_id != -1){
        deliver_udp_data( udp_cbs[cb].sockmandapi_id, 
                          (char*)(udp + 1), 
                          payloadlen, 
                          udp->uh_sport, 
                          ip->ip_src);
    }        
    else{
        struct bufferedpacket *t = malloc(sizeof(struct bufferedpacket));
        t->next = 0;
        t->ip = malloc(ip->ip_len);
        memcpy(t->ip, ip, ip->ip_len);
        
        if(udp_cbs[cb].bufferedpacketsfirst == 0){
            udp_cbs[cb].bufferedpacketsfirst = t;
            udp_cbs[cb].bufferedpacketslast = t;
        }
        
        else{
            udp_cbs[cb].bufferedpacketslast->next = t;
            udp_cbs[cb].bufferedpacketslast = t;
        }
    }
}